/*
 * Copyright (c) 2002-2013, Mairie de Paris
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice
 *     and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice
 *     and the following disclaimer in the documentation and/or other materials
 *     provided with the distribution.
 *
 *  3. Neither the name of 'Mairie de Paris' nor 'Lutece' nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * License 1.0
 */
package fr.paris.lutece.plugins.mvc;

import fr.paris.lutece.plugins.mvc.annotations.Action;
import fr.paris.lutece.plugins.mvc.annotations.Controller;
import fr.paris.lutece.plugins.mvc.annotations.View;
import fr.paris.lutece.portal.service.i18n.I18nService;
import fr.paris.lutece.portal.service.message.SiteMessageException;
import fr.paris.lutece.portal.service.plugin.Plugin;
import fr.paris.lutece.portal.service.security.UserNotSignedException;
import fr.paris.lutece.portal.service.template.AppTemplateService;
import fr.paris.lutece.portal.service.util.AppException;
import fr.paris.lutece.portal.service.util.AppLogService;
import fr.paris.lutece.portal.service.util.AppPropertiesService;
import fr.paris.lutece.portal.web.LocalVariables;
import fr.paris.lutece.portal.web.xpages.XPage;
import fr.paris.lutece.portal.web.xpages.XPageApplication;
import fr.paris.lutece.util.ErrorMessage;
import fr.paris.lutece.util.bean.BeanUtil;
import fr.paris.lutece.util.beanvalidation.BeanValidationUtil;
import fr.paris.lutece.util.html.HtmlTemplate;
import fr.paris.lutece.util.url.UrlItem;

import org.apache.log4j.Logger;

import org.springframework.util.ReflectionUtils;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import javax.validation.ConstraintViolation;


/**
 * MVC XPage Application
 */
@Deprecated
public abstract class MVCApplication implements XPageApplication
{
    private static final String MARK_ERRORS = "errors";
    private static final String MARK_INFOS = "infos";
    private static final String MARK_MESSAGE_BOX = "messageBox";
    private static final String URL_PORTAL = "Portal.jsp";
    private static final String PATH_PORTAL = "jsp/site/";
    private static final String PARAMETER_PAGE = "page";
    private static final String PARAMETER_VIEW = "view";
    private static final String PARAMETER_ACTION = "action";
    private static final String PREFIX_VIEW = "view_";
    private static final String PREFIX_ACTION = "action_";
    private static final String VIEW_MESSAGEBOX = "messageBox";
    private List<ErrorMessage> _listErrors = new ArrayList<ErrorMessage>(  );
    private List<ErrorMessage> _listInfos = new ArrayList<ErrorMessage>(  );
    private MVCMessageBox _messageBox;
    private Controller _controller = getClass(  ).getAnnotation( Controller.class );
    private Logger _logger = Logger.getLogger( "lutece.mvc" );

    /**
     * Returns the content of the page
     *
     * @param request The http request
     * @param nMode The current mode
     * @param plugin The plugin object
     * @return The XPage
     * @throws fr.paris.lutece.portal.service.message.SiteMessageException
     *             Message displayed if an exception occurs
     * @throws UserNotSignedException if an authentication is required by a view
     */
    @Override
    public XPage getPage( HttpServletRequest request, int nMode, Plugin plugin )
        throws SiteMessageException, UserNotSignedException
    {
        return processController( request );
    }

    ////////////////////////////////////////////////////////////////////////////
    // Controller

    /**
     * XPage controller
     * @param request The HTTP request
     * @return The XPage
     * @throws UserNotSignedException if an authentication is required by a view
     * @throws SiteMessageException if a message is thrown by an action
     */
    private XPage processController( HttpServletRequest request )
        throws UserNotSignedException, SiteMessageException
    {
        Method[] methods = ReflectionUtils.getAllDeclaredMethods( getClass(  ) );

        try
        {
            if ( isMessageBox( request ) )
            {
                return messageBox( request );
            }

            // Process views
            Method m = findViewAnnotedMethod( methods, getView( request ) );

            if ( m != null )
            {
                _logger.debug( "MVC controller - process view : '" + getView( request ) + "'" );

                return (XPage) m.invoke( this, request );
            }

            // Process actions
            m = findActionAnnotedMethod( methods, getAction( request ) );

            if ( m != null )
            {
                _logger.debug( "MVC controller - process action : '" + getAction( request ) + "'" );

                return (XPage) m.invoke( this, request );
            }

            // No view or action found so display the default view
            m = findDefaultViewMethod( methods );
            _logger.debug( "MVC controller - process default view" );

            return (XPage) m.invoke( this, request );
        }
        catch ( InvocationTargetException e )
        {
            if ( e.getTargetException(  ) instanceof UserNotSignedException )
            {
                throw (UserNotSignedException) e.getTargetException(  );
            }

            if ( e.getTargetException(  ) instanceof SiteMessageException )
            {
                throw (SiteMessageException) e.getTargetException(  );
            }

            throw new AppException( "MVC Error dispaching view and action ", e );
        }
        catch ( IllegalAccessException e )
        {
            throw new AppException( "MVC Error dispaching view and action ", e );
        }
    }

    /**
     * Get the view parameter
     * @param request The HTTP request
     * @return The view parameter or null if not found
     */
    protected String getView( HttpServletRequest request )
    {
        String strView = request.getParameter( PARAMETER_VIEW );

        if ( strView != null )
        {
            return strView;
        }

        Enumeration<String> parameters = request.getParameterNames(  );

        while ( parameters.hasMoreElements(  ) )
        {
            String strParameter = parameters.nextElement(  );

            if ( strParameter.startsWith( PREFIX_VIEW ) )
            {
                strView = strParameter.substring( PREFIX_VIEW.length(  ) );

                break;
            }
        }

        return strView;
    }

    /**
     * Get the action parameter or any parameter with the prefix 'action_'
     * @param request The HTTP request
     * @return The action parameter or null if not found
     */
    protected String getAction( HttpServletRequest request )
    {
        String strAction = request.getParameter( PARAMETER_ACTION );

        if ( strAction != null )
        {
            return strAction;
        }

        Enumeration<String> parameters = request.getParameterNames(  );

        while ( parameters.hasMoreElements(  ) )
        {
            String strParameter = parameters.nextElement(  );

            if ( strParameter.startsWith( PREFIX_ACTION ) )
            {
                strAction = strParameter.substring( PREFIX_ACTION.length(  ) );

                break;
            }
        }

        return strAction;
    }

    /**
     * Find the method that provide a given view
     * @param methods The methods list
     * @param strView The view
     * @return The method
     */
    private Method findViewAnnotedMethod( Method[] methods, String strView )
    {
        if ( strView != null )
        {
            for ( Method m : methods )
            {
                if ( m.isAnnotationPresent( View.class ) && strView.equals( m.getAnnotation( View.class ).value(  ) ) )
                {
                    return m;
                }
            }

            _logger.warn( "MVC controller - No method found to process view : '" + strView + "'" );
        }

        return null;
    }

    /**
     * Find the method that provide a given action
     * @param methods The methods list
     * @param strAction The action
     * @return The method
     */
    private Method findActionAnnotedMethod( Method[] methods, String strAction )
    {
        if ( strAction != null )
        {
            for ( Method m : methods )
            {
                if ( m.isAnnotationPresent( Action.class ) &&
                        strAction.equals( m.getAnnotation( Action.class ).value(  ) ) )
                {
                    return m;
                }
            }

            _logger.warn( "MVC controller - No method found to process action : '" + strAction + "'" );
        }

        return null;
    }

    /**
     * Find the method that provide the default view
     * @param methods The methods list
     * @return The method
     */
    private Method findDefaultViewMethod( Method[] methods )
    {
        for ( Method m : methods )
        {
            if ( m.isAnnotationPresent( View.class ) && m.getAnnotation( View.class ).defaultView(  ) )
            {
                return m;
            }
        }

        return null;
    }

    /**
     * Returns the XPage name
     * @return The XPage name
     */
    protected String getXPageName(  )
    {
        return _controller.xpageName(  );
    }

    /**
     * Returns the default page title
     * @param locale The locale
     * @return The default page title
     */
    protected String getDefaultPageTitle( Locale locale )
    {
        if ( !_controller.pageTitleProperty(  ).equals( "" ) )
        {
            return AppPropertiesService.getProperty( _controller.pageTitleProperty(  ) );
        }
        else if ( !_controller.pageTitleI18nKey(  ).equals( "" ) )
        {
            return I18nService.getLocalizedString( _controller.pageTitleI18nKey(  ), locale );
        }

        return _controller.xpageName(  );
    }

    /**
     * Returns the default page path
     * @param locale The locale
     * @return The default pagepath
     */
    protected String getDefaultPagePath( Locale locale )
    {
        if ( !_controller.pagePathProperty(  ).equals( "" ) )
        {
            return AppPropertiesService.getProperty( _controller.pagePathProperty(  ) );
        }
        else if ( !_controller.pagePathI18nKey(  ).equals( "" ) )
        {
            return I18nService.getLocalizedString( _controller.pagePathI18nKey(  ), locale );
        }

        return _controller.xpageName(  );
    }

    ////////////////////////////////////////////////////////////////////////////
    // XPage utils

    /**
     * Returns a new XPage object with default values
     * @return An XPage Object
     */
    protected XPage getXPage(  )
    {
        XPage page = new XPage(  );

        page.setTitle( getDefaultPageTitle( Locale.getDefault(  ) ) );
        page.setPathLabel( getDefaultPagePath( Locale.getDefault(  ) ) );

        return page;
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template
     * @param strTemplate The template
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate )
    {
        XPage page = getXPage(  );

        HtmlTemplate t = AppTemplateService.getTemplate( strTemplate );
        page.setContent( t.getHtml(  ) );

        return page;
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template using a default model and for a given locale
     * @param strTemplate The template
     * @param locale The locale
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate, Locale locale )
    {
        return getXPage( strTemplate, locale, getModel(  ) );
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template using a given model and for a given locale
     * @param strTemplate The template
     * @param locale The locale
     * @param model The model
     *
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate, Locale locale, Map<String, Object> model )
    {
        XPage page = getXPage(  );

        HtmlTemplate t = AppTemplateService.getTemplate( strTemplate, locale, model );
        page.setContent( t.getHtml(  ) );
        page.setTitle( getDefaultPageTitle( locale ) );
        page.setPathLabel( getDefaultPagePath( locale ) );

        return page;
    }

    /**
     * Get a model Object filled with default values
     * @return The model
     */
    protected Map<String, Object> getModel(  )
    {
        Map<String, Object> model = new HashMap<String, Object>(  );
        fillCommons( model );

        return model;
    }

    ////////////////////////////////////////////////////////////////////////////
    // Bean processing

    /**
     * Populate a bean using parameters in http request
     * @param bean bean to populate
     * @param request http request
     */
    protected void populate( Object bean, HttpServletRequest request )
    {
        BeanUtil.populate( bean, request );
    }

    /**
     * Validate a bean
     * @param <T> The bean class
     * @param bean The bean
     * @return true if validated otherwise false
     */
    protected <T> boolean validateBean( T bean )
    {
        Set<ConstraintViolation<T>> errors = BeanValidationUtil.validate( bean );

        if ( errors.isEmpty(  ) )
        {
            return true;
        }

        for ( ConstraintViolation<T> constraint : errors )
        {
            MVCMessage error = new MVCMessage(  );
            error.setMessage( constraint.getMessage(  ) );
            _listErrors.add( error );
        }

        return false;
    }

    /**
     * Validate a bean
     * @param <T> The bean class
     * @param bean The bean
     * @param locale The locale
     * @return true if validated otherwise false
     */
    protected <T> boolean validateBean( T bean, Locale locale )
    {
        Set<ConstraintViolation<T>> errors = BeanValidationUtil.validate( bean );

        if ( errors.isEmpty(  ) )
        {
            return true;
        }

        for ( ConstraintViolation<T> constraint : errors )
        {
            MVCMessage error = new MVCMessage(  );
            error.setMessage( I18nService.getLocalizedString( constraint.getMessage(  ), locale ) );
            _listErrors.add( error );
        }

        return false;
    }

    /**
     * Add an error message
     * @param strMessage The message
     */
    protected void addError( String strMessage )
    {
        _listErrors.add( new MVCMessage( strMessage ) );
    }

    /**
     * Add an error message
     * @param strMessageKey The message
     * @param locale The locale
     */
    protected void addError( String strMessageKey, Locale locale )
    {
        _listErrors.add( new MVCMessage( I18nService.getLocalizedString( strMessageKey, locale ) ) );
    }

    /**
     * Add an info message
     * @param strMessage The message
     */
    protected void addInfo( String strMessage )
    {
        _listInfos.add( new MVCMessage( strMessage ) );
    }

    /**
     * Add an info message
     * @param strMessageKey The message key
     * @param locale The locale
     */
    protected void addInfo( String strMessageKey, Locale locale )
    {
        _listInfos.add( new MVCMessage( I18nService.getLocalizedString( strMessageKey, locale ) ) );
    }

    /**
     * Fill the model with commons objects used in templates
     * @param model The model
     */
    protected void fillCommons( Map<String, Object> model )
    {
        List<ErrorMessage> listErrors = new ArrayList<ErrorMessage>( _listErrors );
        List<ErrorMessage> listInfos = new ArrayList<ErrorMessage>( _listInfos );
        model.put( MARK_ERRORS, listErrors );
        model.put( MARK_INFOS, listInfos );
        _listErrors.clear(  );
        _listInfos.clear(  );
    }

    ////////////////////////////////////////////////////////////////////////////
    // Redirect utils

    /**
     * Redirect to requested page
     *
     * @param request the http request
     * @param strTarget the targeted page
     * @return the page requested
     */
    protected XPage redirect( HttpServletRequest request, String strTarget )
    {
        HttpServletResponse response = LocalVariables.getResponse(  );

        try
        {
            _logger.debug( "Redirect :" + strTarget );
            response.sendRedirect( strTarget );
        }
        catch ( IOException e )
        {
            _logger.error( "Unable to redirect : " + strTarget + " : " + e.getMessage(  ), e );
        }

        return new XPage(  );
    }

    /**
     * Redirect to an url defined by given parameters
     * @param request The HTTP request
     * @param strView The View name
     * @param strParameter The additional parameter
     * @param nValue The additional parameter's value
     * @return The XPage redirected
     */
    protected XPage redirect( HttpServletRequest request, String strView, String strParameter, int nValue )
    {
        UrlItem url = new UrlItem( getViewUrl( strView ) );
        url.addParameter( strParameter, nValue );

        return redirect( request, url.getUrl(  ) );
    }

    /**
     * Redirect to an url defined by given parameters
     * @param request The HTTP request
     * @param strView The View name
     * @param strParameter1 The first additional parameter
     * @param nValue1 The first additional parameter's value
     * @param strParameter2 The second additionnal parameter
     * @param nValue2 The second additionnal parameter's value
     * @return The XPage redirected
     */
    protected XPage redirect( HttpServletRequest request, String strView, String strParameter1, int nValue1,
        String strParameter2, int nValue2 )
    {
        UrlItem url = new UrlItem( getViewUrl( strView ) );
        url.addParameter( strParameter1, nValue1 );
        url.addParameter( strParameter2, nValue2 );

        return redirect( request, url.getUrl(  ) );
    }

    /**
     * Redirect to requested view
     *
     * @param request the http request
     * @param strView the targeted view
     * @return the page requested
     */
    protected XPage redirectView( HttpServletRequest request, String strView )
    {
        UrlItem url = new UrlItem( URL_PORTAL );
        url.addParameter( PARAMETER_PAGE, getXPageName(  ) );
        url.addParameter( PARAMETER_VIEW, strView );

        return redirect( request, url.getUrl(  ) );
    }

    /**
     * Get a View URL
     * @param strView The view name
     * @return The URL
     */
    protected String getViewUrl( String strView )
    {
        UrlItem url = new UrlItem( URL_PORTAL );
        url.addParameter( PARAMETER_PAGE, getXPageName(  ) );
        url.addParameter( PARAMETER_VIEW, strView );

        return url.getUrl(  );
    }

    /**
     * Gets the view URL with the JSP path
     * @param strView The view
     * @return The URL
     */
    protected String getViewFullUrl( String strView )
    {
        return PATH_PORTAL + getViewUrl( strView );
    }

    /**
     * Get Action URL
     * @param strAction The view name
     * @return The URL
     */
    protected String getActionUrl( String strAction )
    {
        UrlItem url = new UrlItem( URL_PORTAL );
        url.addParameter( PARAMETER_PAGE, getXPageName(  ) );
        url.addParameter( PARAMETER_ACTION, strAction );

        return url.getUrl(  );
    }

    /**
     * Initiates a file download
     * @param strData Data of the file to download
     * @param strFilename Name of file
     * @param strContentType content type to set to the response
     * @return The page requested
     */
    public XPage download( String strData, String strFilename, String strContentType )
    {
        HttpServletResponse response = LocalVariables.getResponse(  );
        PrintWriter out = null;
        response.setHeader( "Content-Disposition", "attachment; filename=\"" + strFilename + "\";" );
        response.setHeader( "Content-type", strContentType );
        response.addHeader( "Content-Encoding", "UTF-8" );
        response.addHeader( "Pragma", "public" );
        response.addHeader( "Expires", "0" );
        response.addHeader( "Cache-Control", "must-revalidate,post-check=0,pre-check=0" );
        response.setContentType( "enctype=multipart/form-data;charset=UTF-8" );

        try
        {
            out = response.getWriter(  );
            out.print( strData );
        }
        catch ( IOException e )
        {
            AppLogService.error( e.getStackTrace(  ), e );
        }
        finally
        {
            if ( out != null )
            {
                out.close(  );
            }
        }

        return new XPage(  );
    }

    /**
     * Initiates a download of a byte array
     * @param data Data to download
     * @param strFilename Name of the downloaded file
     * @param strContentType Content type to set to the response
     * @return The page requested
     */
    protected XPage download( byte[] data, String strFilename, String strContentType )
    {
        HttpServletResponse response = LocalVariables.getResponse(  );
        OutputStream os;
        response.setHeader( "Content-Disposition", "attachment; filename=\"" + strFilename + "\";" );
        response.setHeader( "Content-type", strContentType );
        response.addHeader( "Content-Encoding", "UTF-8" );
        response.addHeader( "Pragma", "public" );
        response.addHeader( "Expires", "0" );
        response.addHeader( "Cache-Control", "must-revalidate,post-check=0,pre-check=0" );
        response.setContentType( "enctype=multipart/form-data;charset=UTF-8" );

        try
        {
            os = response.getOutputStream(  );
            os.write( data );
            os.close(  );
        }
        catch ( IOException e )
        {
            AppLogService.error( e.getStackTrace(  ), e );
        }

        return new XPage(  );
    }

    ////////////////////////////////////////////////////////////////////////////
    // MESSAGE BOX MANAGEMENT

    /**
     * Redirect to a Message Box page
     * @param request The HTTP request
     * @param messageBox The MessageBox infos
     * @return A redirect XPage
     */
    protected XPage redirectMessageBox( HttpServletRequest request, MVCMessageBox messageBox )
    {
        _messageBox = messageBox;

        return redirectView( request, VIEW_MESSAGEBOX );
    }

    /**
     * Check if a message box is asked for
     * @param request The HTTP request
     * @return If a message box is asked
     */
    private boolean isMessageBox( HttpServletRequest request )
    {
        String strView = request.getParameter( PARAMETER_VIEW );

        return ( ( strView != null ) && ( strView.equals( VIEW_MESSAGEBOX ) ) );
    }

    /**
     * Default getLocale() implementation. Could be overriden
     * @param request The HTTP request
     * @return The Locale
     */
    protected Locale getLocale( HttpServletRequest request )
    {
        return request.getLocale(  );
    }

    /**
     * Display the Message BOX
     * @param request The HTTP request
     * @return The message box
     */
    private XPage messageBox( HttpServletRequest request )
    {
        _messageBox.localize( getLocale( request ) );

        Map<String, Object> model = getModel(  );
        model.put( MARK_MESSAGE_BOX, _messageBox );

        return getXPage( _messageBox.getTemplate(  ), getLocale( request ), model );
    }
}
